import sys
import socket
import os
import threading
import urllib2
import cookielib
import copy
from ..sysconf import *


def urlopenWithRetry(opener, request, timeout=URL_OPEN_TIMEOUT, retryTime=URL_OPEN_RETRY_TIME):
    if isinstance(request, urllib2.Request):
        url = request.get_full_url()
    else:
        url = request

    while retryTime >= 0:
        try:
            return opener.open(request, timeout=timeout)
        except socket.timeout, e1:
            retryTime -= 1
            # syslog("http timeout, retry again.(retry remains %s) url=%s" % (retryTime, url), LOG_ERROR)
            continue
        except urllib2.HTTPError, error:
            if error.code == 404:
                syslog("http error 404 not found, url=%s" % (url), LOG_ERROR)
                return None
        except Exception, e2:
            syslog(str(e2) + " (retry remains %s)" % (retryTime) + ", at url=" + url, LOG_ERROR)
            retryTime -= 1
            continue
    return None


def resReadWithRetry(opener, request, response, timeout=URL_OPEN_TIMEOUT, retryTime=URL_OPEN_RETRY_TIME):
    if isinstance(request, urllib2.Request):
        url = request.get_full_url()
    else:
        url = request
    needRetry = False
    if response == None:
        return None
    while retryTime >= 0:
        try:
            if needRetry == False:
                return response.read()
            else:
                return opener.open(request, timeout=timeout).read()
        except socket.timeout, e1:
            retryTime -= 1
            # syslog("response read http timeout, retry again.(retry remains %s), url=%s" % (retryTime, url), LOG_ERROR)
            needRetry = True
            continue
        except Exception, e2:
            syslog(str(e2) + " (retry remains %s)" % (retryTime) + ", at url=" + url, LOG_ERROR)
            retryTime -= 1
            needRetry = True
            continue
    return None


def urlReadWithRetry(opener, request, timeout=URL_OPEN_TIMEOUT, retryTime=URL_OPEN_RETRY_TIME):
    return resReadWithRetry(opener, request, urlopenWithRetry(opener, request, timeout, retryTime), timeout, retryTime)


def upDiv(a, b):
    return (a + b - 1) / b


def getShardingConf(size, maxThreadCount=MAX_THREAD_COUNT, perferedSplitNum=SPLIT_NUM):
    tmpCount = upDiv(size, perferedSplitNum)
    if tmpCount > maxThreadCount:
        tmpCount = maxThreadCount

    splitNum = upDiv(size, tmpCount)
    downloadCount = tmpCount
    return [splitNum, downloadCount]


def getPathWithSep(path):
    if path[len(path) - 1] != os.sep:
        return path + os.sep
    else:
        return path


def createDir(savePath):
    # create dir
    if os.path.isdir(savePath) is not True:
        os.makedirs(savePath)


# following from Python cookbook, #475186
def has_colours(stream):
    if not hasattr(stream, "isatty"):
        return False
    if not stream.isatty():
        return False  # auto color only on TTYs
    try:
        import curses

        curses.setupterm()
        return curses.tigetnum("colors") > 2
    except:
        # guess false in case of error
        return False


has_colours = has_colours(sys.stdout)

#constant
BLACK, RED, GREEN, YELLOW, BLUE, MAGENTA, CYAN, WHITE = range(8)
LOG_DEBUG, LOG_INFO, LOG_WARNING, LOG_ERROR = range(4)
level_str = ["DEBUG", "INFO", "WARNING", "ERROR"]


def getLogColour(level):
    color_range = [WHITE, GREEN, YELLOW, RED]
    return color_range[level]


syslogLock = threading.Lock()


def syslog(text, level=LOG_INFO):
    syslogLock.acquire()
    if has_colours:
        seq = "\x1b[1;%dm" % (30 + getLogColour(level)) + "[%s]: " % (level_str[level]) + "\x1b[0m"
        sys.stdout.write(seq)
        sys.stdout.write(text + "\n")
    else:
        sys.stdout.write(text + "\n")
    syslogLock.release()


def dynamicImport(name):
    mod = __import__(name)
    components = name.split('.')
    for comp in components[1:]:
        mod = getattr(mod, comp)
    return mod


#default only copy opener's headers and cookiejar'
def openerDefaultClone(oldOpener, oldCookieJar=None):
    for handler in oldOpener.handlers:
        if isinstance(handler, urllib2.HTTPCookieProcessor):
            oldCookieJar = handler.cookiejar
    newCookieJar = cookielib.CookieJar()
    if oldCookieJar != None:
        newCookieJar._cookies = copy.deepcopy(oldCookieJar._cookies)
        newCookieJar._policy = copy.deepcopy(oldCookieJar._policy)
    newOpener = urllib2.build_opener(urllib2.HTTPCookieProcessor(newCookieJar))
    newOpener.addheaders = copy.deepcopy(oldOpener.addheaders)
    return newOpener



import urllib

from pyquery import PyQuery as pq

from pageBasedHandler import *
from ..common.utils import *


def getUrlFromId(memberID):
    url = 'http://www.pixiv.net/member_illust.php?id=' + str(memberID)
    return url


class PixivHandler(PageBasedHandler):
    def getPageUrl(self, opener, conf):
        urlList = list()
        htmlContent = urlReadWithRetry(opener, conf["url"])
        if htmlContent == None:
            syslog("pixiv plugin pageCrawl init fail! timeout retry too many times, url=%s" % (conf["url"]), LOG_ERROR)
            sys.exit(1)
        d = pq(htmlContent)
        pageLi = d(".page-list").eq(0)("li")
        size = pageLi.size()
        if size == 0:
            size = 1
        syslog("total " + str(size) + " pages", LOG_INFO)
        urlList.append(conf["url"])
        for page in range(2, size + 1):
            urlList.append(conf["url"] + "&p=" + str(page))
        return urlList

    def getPictureUrl(self, pageUrl, opener):
        urlList = list()
        htmlContent = urlReadWithRetry(opener, pageUrl)
        if htmlContent == None:
            syslog("pixiv plugin pictureCrawl init fail! timeout retry too many times, url=%s" % (baseUrl), LOG_ERROR)
            return

        d = pq(htmlContent)
        profile = d(".profile-unit")
        src = profile.find("img").attr("src")
        userName = src[src.find("profile") + len("profile") + 1: src.rfind("/")]
        item = d(".image-item")
        for index in range(0, item.size()):
            href = item.eq(index).find("a").attr("href")
            imgID = href[href.rfind("=") + 1:]
            bigImgUrl = "http://www.pixiv.net/member_illust.php?mode=big&illust_id=" + imgID
            syslog("opening " + bigImgUrl, LOG_INFO)
            bigImgContent = urlReadWithRetry(opener, bigImgUrl)
            if bigImgContent == None:
                syslog("retry too many times at url=" + bigImgUrl, LOG_ERROR)

            bigD = pq(bigImgContent)
            imgUrl = bigD("img").attr("src")
            if imgUrl == None:
                syslog("error!" + " big image url wrong url=" + bigImgUrl, LOG_ERROR)
            else:
                urlList.append(imgUrl)
        return urlList

    def initOpener(self, conf):
        cookieJar = cookielib.CookieJar()
        cookie = urllib2.HTTPCookieProcessor(cookieJar)
        opener = urllib2.build_opener(cookie)
        opener.addheaders = [
            ('Referer', 'http://www.pixiv.net/member_illust.php?mode=medium'),
            ('User-Agent',
             'Mozilla/5.0 (Windows NT 6.2; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/33.0.1750.146 Safari/537.36')]
        postData = {
            "mode": "login",
            "return_to": "/",
            "pixiv_id": conf["pixivId"],
            "pass": conf["password"]
        }
        login_headers = {
            'Referer': 'https://www.secure.pixiv.net/login.php',
            'User-Agent': 'Mozilla/5.0 (Windows NT 6.2; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/33.0.1750.146 Safari/537.36'
        }
        request = urllib2.Request("https://www.secure.pixiv.net/login.php", urllib.urlencode(postData),
                                  headers=login_headers)
        syslog("www.pixiv.net logining!", LOG_INFO)
        urlopenWithRetry(opener, request)
        syslog("login completed!", LOG_INFO)
        self.opener = opener
        return opener

    def initPara(self, parser):
        parser.add_argument('authorId', help='the author id you want to crawl')
        parser.add_argument('savePath', help='the path where the imgs ars saved')
        parser.add_argument('pixivId', help='your pixiv login id')
        parser.add_argument('password', help='your pixiv login password')
        args = parser.parse_args()
        return {
            "url": getUrlFromId(args.authorId),
            "savePath": args.savePath,
            "pixivId": args.pixivId,
            "password": args.password,
            "useRangeHeaders": True
        }



from pyquery import PyQuery as pq

from pageBasedHandler import *
from ..common.utils import *


class GamerskyHandler(PageBasedHandler):
    # def getPageUrl(self, baseUrl, opener, beginPage, endPage):
    def getPageUrl(self, opener, paraConf):
        urlList = list()
        for page in range(paraConf["beginPage"], paraConf["endPage"] + 1):
            if page == 1:
                crawlUrl = paraConf["url"]
            else:
                crawlUrl = paraConf["url"][0: paraConf["url"].index('.shtml')] + "_" + str(page) + ".shtml"
            urlList.append(crawlUrl)
        return urlList

    def getPictureUrl(self, pageUrl, opener):
        urlList = list()
        htmlContent = urlReadWithRetry(opener, pageUrl)
        if htmlContent == None:
            syslog("GameskyPicIterator init fail at url=" + pageUrl, LOG_ERROR)
        q = pq(htmlContent)
        imgP = q('p')
        for i in range(0, imgP.size()):
            try:
                #check if the sub element has a href attribute
                href = imgP.eq(i).find('a').attr('href')
                if (href == None):
                    continue
                #check postfix 
                imgUrl = href.split('?')[1]
                urlList.append(imgUrl)
                syslog("imgurl find, url=%s" % (imgUrl))
            except Exception, e:
                syslog("Exception:" + Exception + ", " + e, LOG_ERROR)
                continue

        return urlList

    def initOpener(self, conf):
        self.opener = urllib2.build_opener()
        return self.opener

    def initPara(self, parser):
        parser.add_argument('url', help='your url to crawl')
        parser.add_argument('savePath', help='the path where the imgs ars saved')
        parser.add_argument('beginPage', help='the page where we start crawling', type=int)
        parser.add_argument('endPage', help='the page where we end crawling', type=int)
        args = parser.parse_args()
        return {
            "url": args.url,
            "savePath": args.savePath,
            "beginPage": args.beginPage,
            "endPage": args.endPage,
            "useRangeHeaders": True
        }



import time

from rangedownloader import RangeDownloader
from utils import *
from ..sysconf import *


class PictureThread(threading.Thread):
    def __init__(self,
                 startIndex,
                 splitNum,
                 opener,
                 urlList,
                 savePath,
                 useRangeHeaders,
                 downloadLock,
                 handlerName):
        threading.Thread.__init__(self)
        self.startIndex = startIndex
        self.urlList = urlList
        self.opener = opener
        self.splitNum = splitNum
        self.name = handlerName
        self.savePath = savePath
        self.useRangeHeaders = useRangeHeaders
        self.downloadLock = downloadLock

    def run(self):
        syslog(self.getName() + " picture thread start!", LOG_INFO)
        pictureSize = len(self.urlList)
        maxIndex = self.startIndex + self.splitNum - 1
        if maxIndex > pictureSize:
            maxIndex = pictureSize

        for index in range(self.startIndex - 1, maxIndex):
            self.downloadLock.acquire()
            self.__crawl(self.urlList[index], index)
            self.downloadLock.release()

    def __crawl(self, url, picIndex):
        syslog("downloading " + url, LOG_INFO)
        savePath = self.savePath + str(picIndex) + '.jpg'
        if self.useRangeHeaders:
            rd = RangeDownloader(self.opener)
            rd.rangeDownload(url, savePath)
        else:
            content = urlReadWithRetry(self.opener, url)
            if content == None:
                syslog("error! retry too many times, url=" % (url), LOG_ERROR)
                return
            file = open(savePath, "wb")
            file.write(content)
            file.close()


class CrawlerManager(object):
    # conf.url and conf.savePath are required
    def __init__(self, opener, urlList, savePath, useRangeHeaders, maxDownloadCount=MAX_DOWNLOAD_COUNT):
        self.savePath = getPathWithSep(savePath)
        self.useRangeHeaders = useRangeHeaders
        self.opener = opener
        self.urlList = urlList
        self.downloadLock = threading.Semaphore(maxDownloadCount)
        createDir(self.savePath)

    def startCrawl(self):
        beginTime = time.time()
        self.PictureThreadList = list()
        index = 1
        count = 1
        picNum = len(self.urlList)
        splitNum = getShardingConf(picNum)[0]

        while index <= picNum:
            picThread = PictureThread(
                index,
                splitNum,
                self.opener,
                self.urlList,
                self.savePath,
                self.useRangeHeaders,
                self.downloadLock,
                "pictureThread:" + str(count))
            self.PictureThreadList.append(picThread)
            picThread.start()
            index += splitNum
            count += 1

        for th in self.PictureThreadList:
            th.join()

        endTime = time.time()
        cost = round(endTime - beginTime, 1)
        average = cost / picNum
        syslog("download complete!", LOG_INFO)
        syslog('crawl ' + str(picNum) + " imgs", LOG_INFO)
        syslog('total cost: ' + str(cost) + 's', LOG_INFO)
        syslog('average ' + str(average) + 's per image', LOG_INFO)

    def stopCrawl(self):
        #TODO not surported yet
        pass



from baseHandler import *
from ..common.utils import *
from ..third_party.abc import abstractmethod


class PictureUrlThread(threading.Thread):
    def __init__(self, threadName, pageUrl, handler):
        threading.Thread.__init__(self, name=threadName)
        self.pageUrl = pageUrl
        self.handler = handler

    def run(self):
        syslog("getting page url=%s" % (self.pageUrl), LOG_INFO)
        self.pictureUrlList = self.handler.getPictureUrl(self.pageUrl, copy.copy(self.handler.opener))

    def getPictureUrlList(self):
        return self.pictureUrlList


# abstruct base class
class PageBasedHandler(BaseHandler):
    def getUrlList(self, conf):
        pageUrlList = self.getPageUrl(copy.copy(self.opener), conf)
        pageNum = len(pageUrlList)
        threadList = list()
        for index, url in enumerate(pageUrlList):
            th = PictureUrlThread("pictureUrlThread:%s" % (index), url, self)
            threadList.append(th)
            th.start()

        pictureUrlList = list()
        for th in threadList:
            th.join()
            pictureUrlList.extend(th.getPictureUrlList())

        return pictureUrlList


    @abstractmethod
    def getPageUrl(self, opener, paraConf):
        pass

    @abstractmethod
    def getPictureUrl(self, pageUrl, opener):
        pass



from utils import *
from ..sysconf import *


class RangeDownloadThread(threading.Thread):
    def __init__(self, startByte, opener, request, file, lock):
        threading.Thread.__init__(self)
        self.startByte = startByte
        self.request = request
        self.opener = openerDefaultClone(opener)
        self.file = file
        self.lock = lock

    def run(self):
        rangeContent = self.__checkRangeHeaderSupport()
        self.lock.acquire()
        self.file.seek(self.startByte)
        if rangeContent is not None:
            self.file.write(rangeContent)
            self.ret = 0  # download succeed
        else:
            syslog("download fail in range:%s, file=%s" % (self.startByte, self.file.name), LOG_ERROR)
            self.ret = -1  # download fail
        self.lock.release()

    def __checkRangeHeaderSupport(self):
        retryTime = 10
        while retryTime >= 0:
            res = urlopenWithRetry(self.opener, self.request)
            if res == None:
                return None
            if res.info().get("Content-Range") != None:
                try:
                    return res.read()
                except Exception, e:
                    retryTime -= 1
                    continue

            retryTime -= 1

        if retryTime == -1:
            syslog("rangeHeader not support at this file part, file=%s, url=%s, startByte=%s" % (
                self.file.name, self.request.get_full_url(), self.startByte), LOG_ERROR)
        return None


class RangeDownloader(object):
    def __init__(self, opener):
        self.opener = openerDefaultClone(opener)
        self.downloadThreadList = list()
        self.fileList = list()
        self.lock = threading.Lock()

    def rangeDownload(self, url, savePath, partNum=RANGE_PART_NUM):
        self.savePath = savePath
        self.url = url
        fileSize = self.__getFileSize(url)
        if fileSize == 0:
            return
        shardingSize = self.__sharding(fileSize, partNum)
        beginByte = 0
        while beginByte < fileSize:
            file = open(savePath, "wb")
            self.fileList.append(file)
            endByte = beginByte + shardingSize - 1
            if (endByte >= fileSize):
                endByte = fileSize - 1
            req = urllib2.Request(url)
            req.headers['Range'] = 'bytes=%s-%s' % (beginByte, endByte)
            th = RangeDownloadThread(beginByte, self.opener, req, file, self.lock)
            self.downloadThreadList.append(th)
            th.start()
            beginByte += shardingSize

        self.__handleResult()

    def __handleResult(self):
        hasError = False
        for th in self.downloadThreadList:
            th.join()

        for th in self.downloadThreadList:
            if th.ret != 0:
                hasError = True

        for f in self.fileList:
            f.close()

        if hasError:
            syslog("download fail file=%s, url=%s" % (self.savePath, self.url), LOG_ERROR)
            os.remove(self.savePath)

    def __getFileSize(self, url):
        res = urlopenWithRetry(self.opener, url)
        if res == None:
            return 0
        return int(res.info()['Content-Length'])

    def __sharding(self, fileSize, partNum):
        return upDiv(fileSize, partNum)
